# -*- coding: utf-8 -*-
"""workerpool"""

__author__  = "Adrien DELLE CAVE <adc@doowan.net>"
__license__ = """
    Copyright (C) 2015  doowan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""


import logging
import Queue
import threading
import time


LOG = logging.getLogger('sonicprobe.workerpool')


class WorkerExit(object):
    pass


class WorkerThread(threading.Thread):
    def __init__(self, pool):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.pool       = pool
        self.life_time  = None

    def start(self):
        self.life_time  = time.time()
        return threading.Thread.start(self)

    def expired(self):
        if self.pool.life_time > 0 \
           and self.life_time > 0 \
           and (time.time() - self.life_time) >= self.pool.life_time:
            return True

        return False

    def run(self):
        while True:
            task = self.pool.tasks.get()
            if isinstance(task, WorkerExit):
                break

            self.pool.count_lock.acquire()
            self.pool.working += 1
            if (self.pool.working >= self.pool.workers) \
               and (self.pool.workers < self.pool.max_workers):
                self.pool.add()
            self.pool.count_lock.release()

            func, cb, args, kargs = task
            try:
                ret = func(*args, **kargs)
                if cb:
                    cb(ret)
            except Exception, e:
                LOG.exception("Unexpected error: %r", e)

            self.pool.count_lock.acquire()
            self.pool.working -= 1
            self.pool.count_lock.release()

            if self.expired():
                break

        self.pool.count_lock.acquire()
        self.pool.workers -= 1
        if not self.pool.workers:
            self.pool.kill_event.set()
        self.pool.count_lock.release()


class WorkerPool(object):
    def __init__(self, queue = None, max_workers = 10, life_time = None):
        self.tasks          = queue or Queue.Queue()
        self.workers        = 0
        self.working        = 0
        self.max_workers    = max_workers
        self.life_time      = life_time
        self.kill_event     = threading.Event()
        self.count_lock     = threading.RLock()

        self.kill_event.set()

    def kill(self, nb = 1):
        """
        Kill one or many workers.
        """
        self.count_lock.acquire()
        if nb > self.workers:
            nb = self.workers
        self.count_lock.release()
        for x in xrange(nb):
            self.tasks.put(WorkerExit())

    def set_max_workers(self, nb):
        """
        Set the maximum workers to create.
        """
        self.count_lock.acquire()
        self.max_workers = nb
        if self.workers > self.max_workers:
            self.kill(self.workers - self.max_workers)
        self.count_lock.release()

    def add(self, nb = 1):
        """
        Create one or many workers.
        """
        for x in xrange(nb):
            self.count_lock.acquire()
            self.workers   += 1
            self.kill_event.clear()
            self.count_lock.release()
            WorkerThread(self).start()

    def run(self, target, callback = None, *args, **kargs):
        """
        Start task.
        @target: callable to run with *args and **kargs arguments.
        @callback: callable executed after target.
        """
        self.count_lock.acquire()
        if not self.workers:
            self.add()
        self.count_lock.release()
        self.tasks.put((target, callback, args, kargs))

    def killall(self, wait = None):
        """
        Kill all active workers.
        @wait: Seconds to wait until last worker ends.
               If None it waits forever.
        """
        self.count_lock.acquire()
        self.kill(self.workers)
        self.count_lock.release()
        self.kill_event.wait(wait)
