# -*- coding: utf-8 -*-
"""keystore"""

__author__  = "Adrien DELLE CAVE <adc@doowan.net>"
__license__ = """
    Copyright (C) 2017  doowan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import logging
import time
import threading

LOG = logging.getLogger('sonicprobe.keystore')


class Keystore(object):
    def __init__(self):
        self.__data     = {}
        self.__lock     = {}
        self.__updated  = {}

    def add(self, name, lock = False):
        if not self.__data.has_key(name):
            self.__lock[name]       = threading.RLock()
            if lock:
                self.__lock[name].acquire()
            self.__data[name]       = {}
            self.__updated[name]    = 0
            if lock:
                self.__lock[name].release()
        return self

    def get(self, name, key, default = None, lock = False):
        if not self.__lock.has_key(name):
            return default

        if lock:
            self.__lock[name].acquire()
        r = self.__data[name].get(key, default)
        if lock:
            self.__lock[name].release()
        return r

    def set(self, name, key, value = None, lock = False):
        if not self.__lock.has_key(name):
            self.add(name)
        if lock:
            self.__lock[name].acquire()
        self.__data[name][key]  = value
        self.__updated[name]    = time.time()
        if lock:
            self.__lock[name].release()
        return self

    def exists(self, name, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock.acquire()

        r = self.__data.has_key(name)

        if lock and self.__lock.has_key(name):
            self.__lock.release()
        return r

    def iteritems(self, name):
        for item in self.__data[name].items():
            yield item

    def iterkeys(self, name):
        for key in self.__data[name].keys():
            yield key

    def itervalues(self, name):
        for value in self.__data[name].values():
            yield value

    def has_key(self, name, key, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()

        r = self.__data[name].has_key(key)

        if lock and self.__lock.has_key(name):
            self.__lock[name].release()
        return r

    def delete(self, name, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()

        if self.__data.has_key(name):
            del self.__data[name]
        if self.__updated.has_key(name):
            del self.__updated[name]

        if lock and self.__lock.has_key(name):
            self.__lock[name].release()

        if self.__lock.has_key(name):
            del self.__lock[name]
        return self

    def remove(self, name, key, lock = False):
        if lock:
            self.__lock[name].acquire()
        if self.__data.has_key(key):
            del self.__data[name][key]
        self.__updated[name] = time.time()
        if lock:
            self.__lock[name].release()
        return self

    def updated_at(self, name, default = None, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()
        r = self.__updated.get(name, default)
        if lock and self.__lock.has_key(name):
            self.__lock[name].release()
        return r

    def updated_delta(self, name, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()
        r = time.time() - self.__updated[name]
        if lock and self.__lock.has_key(name):
            self.__lock[name].release()
        return r

    def expired(self, name, expire, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()

        try:
            if not self.__updated.has_key(name):
                return None
            if expire < 0:
                return False
            return self.updated_delta(name) > expire
        finally:
            if lock and self.__lock.has_key(name):
                self.__lock[name].release()

    def purge(self, name, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()

        if self.__data.has_key(name):
            self.__data[name]       = {}
        if self.__updated.has_key(name):
            self.__updated[name]    = 0

        if lock and self.__lock.has_key(name):
            self.__lock[name].release()
        return self

    def reset(self, name, lock = False):
        if lock and self.__lock.has_key(name):
            self.__lock[name].acquire()

        if self.__data.has_key(name):
            self.__data[name]       = {}
        if self.__updated.has_key(name):
            self.__updated[name]    = time.time()

        if lock and self.__lock.has_key(name):
            self.__lock[name].release()
        return self

    def acquire(self, name, blocking = True, lock = True):
        self.add(name, lock)
        return self.__lock[name].acquire(blocking)

    def try_acquire(self, name, timeout = None, exists = False):
        if timeout is not None:
            endtime = time.time() + timeout

        if not exists:
            self.add(name, False)
        elif not self.exists(name, False):
            return False

        while True:
            if not self.__lock.has_key(name):
                return False
            elif self.__lock[name].acquire(False):
                return True
            elif timeout is not None:
                remaining = endtime - time.time()
                if remaining <= 0:
                    return None
            else:
                return False

    def release(self, name):
        if self.__lock.has_key(name):
            self.__lock[name].release()
        return self

    def try_release(self, name):
        try:
            self.release(name)
        except RuntimeError:
            pass

        return self

    def list(self):
        for key in self.__data.keys():
            yield key
